package ca.ubc.cs.elide.nodes;

import ca.ubc.cs.elide.*;
import ca.ubc.cs.elide.transformations.*;
import ca.ubc.cs.elide.visitors.*;
import ca.ubc.cs.elide.utils.*;
import java.util.*;
import java.lang.reflect.*;

/**
 * <code>ChildBearingNode</code> is an abstract class representing
 * nodes which can act as ancestors of child nodes in the AST.
 */
public abstract class ChildBearingNode extends ParseNode
{
	private Map indexMap = new HashMap();
	private List children = new ArrayList();
	
	public ChildBearingNode() {}
	public ChildBearingNode(String name) { super(name); }
		
	/**
	 * Adds a child to the children of this <code>ChildBearingNode</code>
	 */
	public void add(ParseNode node)
	{		
		add(children.size(), node);
	}
	
	/**
	 * Adds a child to the children of this <code>ChildBearingNode</code>,
	 * at the index specified.
	 */
	public void add(int index, ParseNode node)
	{
		node.setParent(this);
		children.add(index, node);
		indexChild(node);
	}		
	
	/**
	 * Adds a transformation to this <code>ChildBearingNode</code>, and
	 * registers this transformation.
	 *
	 * Has the effect of adding the keyword associated with <code>node</code> to
	 * the current node in the base code.
	 */
	public void addTransform(TransformNode node)
	{
		add(node);
		TransformCollector collector = getRoot().transformCollector;
		collector.addTransform(node);
	}
	
	public void remove(ParseNode node)
	{
		if(children.contains(node))
		{
			children.remove(node);
			//FIXME - remove from indexMap too
		}
	}
	
	protected void indexChild(ParseNode node)
	{
		Map index = (Map) indexMap.get(node.getClass());
		if(index == null)
		{
			index = new HashMap();
			indexMap.put(node.getClass(), index);
		}
		index.put(node.name, node);
	}

	/**
	 * Returns all <code>ParseNode</code>s which are children of this <code>ChildBearingNode</code>,
	 * as a <code>List</code>.
	 */
	public List getChildren()
	{
		return children;
	}
	
	protected ParseNode getChild(Class type, String name)
	{
		Map index = (Map) indexMap.get(type);
		if(index != null)
			return (ParseNode) index.get(name);
		else
			return null;
	}
	
	protected ParseNode findChild(Class type, String name)
	{
		ParseNode child = getChild(type, name);
		if(child != null)
			return child;
		
		Iterator i = getChildren().iterator();
		while(i.hasNext())
		{
			ParseNode node = (ParseNode) i.next();
			if(node instanceof ChildBearingNode)
			{
				child = ((ChildBearingNode) node).findChild(type, name);
				if(child != null)
					return child;
			}
		}
		
		return null;
	}
	
	/**
	 * Returns all <code>ParseNode</code>s which are children of this <code>ChildBearingNode</code>,
	 * as an <code>Array</code>.
	 */
	protected Object[] getChildren(Class type)
	{
		Iterator i = children.iterator();
		List result = new ArrayList();
		while(i.hasNext())
		{
			Object child = i.next();
			if(type.isInstance(child))
				result.add(child);
		}
		
		Object[] sample = null;
		try { sample = (Object[]) Array.newInstance(type, 0); }
		catch(Exception e){ }
		
		return result.toArray(sample);
	}
	
	/**
	 * Returns a deep copy of this <code>ChildBearingNode</code> object.
	 */
	public Object clone()
	{
		return InternalUtils.cloneObject(this);
	}
}

