package ca.ubc.cs.elide.nodes;

import ca.ubc.cs.elide.*;
import ca.ubc.cs.elide.utils.*;
import java.util.*;

/**
 * <code>ClassNode</code> objects represent classes in the target system.
 * The easiest way to create <code>ClassNode</code>s is to call {@link NodeFactory#createClass}
 */
public class ClassNode extends ModifierBearingNode implements CodeNode
{	
	boolean isInterface = false;	
	String superclass = "Object";
	
	public ClassNode() {}
	public ClassNode(String name) { super(name); }
		
	/**
	 * Returns the fully-qualified name of the entity (class, interface) represented by this <code>ClassNode</code>  object, as a <code>String</code>.
	 */
	public String getName() { return getPackage().getName() + "." + name; }

	/**
	 * Returns the unqualified name of the entity (class, interface) represented by this <code>ClassNode</code>  object, as a <code>String</code>.
	 */
	public String getShortName() { return name; }
	
	/**
	 * Returns an array of <code>MethodNode</code> objects reflecting all the methods declared by the class or interface represented by this 
	 * <code>ClassNode</code> object
	 */
	public MethodNode[] getDeclaredMethods() { return (MethodNode[]) getChildren(MethodNode.class); }
	
	/**
	 * Returns an array of <code>FieldNode</code> objects reflecting all the fields declared by the class or interface represented by this <code>ClassNode</code> object.
	 */
	public FieldNode[] getDeclaredFields() { return (FieldNode[]) getChildren(FieldNode.class); }

	/**
	 * Returns an array of<code> ClassNode</code> objects reflecting all the classes and interfaces declared as members of the class represented by 
	 * this <code>ClassNode</code> object.
	 */
	public ClassNode[] getDeclaredClasses() { return (ClassNode[]) getChildren(ClassNode.class); }

	/**
	 * Returns an array of <code>StaticBlockNode</code> objects reflecting all the static blocks of the class represented by this<code> ClassNode</code> object.
	 */
	public StaticBlockNode[] getStaticBlocks() { return (StaticBlockNode[]) getChildren(StaticBlockNode.class); }

	/**
	 * Returns an array of <code>BlockNode</code> objects reflecting all the static blocks of the class represented by this <code>ClassNode</code> object.
	 */
	public InstanceBlockNode[] getBlocks() { return (InstanceBlockNode[]) getChildren(InstanceBlockNode.class); }

	/**
	 *  Determines the interfaces implemented by the class or interface represented by this object.
	 */
	public TypeNode[] getInterfaces() { return (TypeNode[]) getChildren(TypeNode.class); }	
	
	/**
	 * Returns an array of <code>ImportNode</code> objects reflecting all units imported to the class represented by this <code>ClassNode</code> object.
	 */
	public ImportNode[] getImports() { return (ImportNode[]) getChildren(ImportNode.class); }	

	/**
	 * Returns an array of <code>ConstructorNode</code> objects reflecting all the fields declared by the class or interface represented by this <code>ClassNode</code> object.
	 */
	public ConstructorNode[] getDeclaredConstructors() { return (ConstructorNode[]) getChildren(ConstructorNode.class); }
	
	/**
	 * Returns the <code>ClassNode</code> representing the superclass of the entity (class, interface) represented by this <code>ClassNode</code>.
	 */
	public String getSuperclass() { return superclass; }
	
	/**
	 * Sets the  superclass of the entity (class, interface) represented by this <code>ClassNode</code> to the class specified by the given <code>String</code>.
	 */
	public void setSuperclass(String superclass) { this.superclass = superclass; }
		
	/**
	 * Determines if this <code>ClassNode</code> object represents an interface type.
	 */
	public boolean isInterface() { return isInterface; }
	
	/**
	 * Specifies that this <code>ClassNode</code> object represents an interface type.
	 */
	public void setInterface(boolean isInterface) { this.isInterface = isInterface; }
	
	/**
	 * Returns a Method object that reflects the specified declared method of the class or interface represented by this <code>ClassNode</code> object.
	 */
	public MethodNode getDeclaredMethod(String name) { return (MethodNode) getChild(MethodNode.class, name); }

	/**
	 * Returns a <code>FieldNode</code> object that reflects the specified declared field of the class or interface represented by this <code>ClassNode</code>  object.
	 */
	public FieldNode getDeclaredField(String name) { return (FieldNode) getChild(FieldNode.class, name); }

	/**
	 * Returns a <code>ClassNode</code> object reflecting the specified class or interface declared as a member of the class represented by 
	 * this <code>ClassNode</code> object.
	 */
	public ClassNode getDeclaredClass(String name) { return (ClassNode) getChild(ClassNode.class, name); }
	
	/**
	 * Returns a <code>MethodNode</code> object that reflects the specified public member method of the class or interface represented by this <code>ClassNode</code> object.
	 */
	public MethodNode getMethod(String name) { return (MethodNode) getAllChild(MethodNode.class, name); }

	/**
	 * Returns a <code>FieldNode</code> object that reflects the specified public member field of the class or interface represented by this <code>ClassNode</code> object.
	 */
	public FieldNode getField(String name) { return (FieldNode) getAllChild(FieldNode.class, name); }

	/**
	 * Returns a <code>ClassNode</code> object that reflects the specified inner class or interface of the class represented by this <code>ClassNode</code> object.
	 */
	public ClassNode getClass(String name) { return (ClassNode) getAllChild(ClassNode.class, name); }

	/**
	 *  Returns an array containing <code>FieldNode</code> objects reflecting all the accessible public fields of the class or interface represented by this <code>ClassNode</code> object.
	 */
	public FieldNode[] getFields() { return (FieldNode[]) getAllChildren(FieldNode.class); }

	/**
	 * Returns an array containing <code>MethodNode</code> objects reflecting all the public member methods of the class or interface 
	 * represented by this <code>ClassNode</code> object, including those declared by the class or interface and and those inherited from superclasses and 
	 * superinterfaces.
	 */
	public MethodNode[] getMethods() { return (MethodNode[]) getAllChildren(MethodNode.class); }

	/**
	 * Returns an array containing <code>ClassNode</code> objects representing all the public classes and interfaces that are members 
	 * of the class represented by this <code>ClassNode</code> object.
	 */
	public ClassNode[] getClasses() { return (ClassNode[]) getAllChildren(ClassNode.class); }
	
	private Object[] getAllChildren(Class type)
	{
			Object[] children = getChildren(type);
			ClassNode superClassNode = resolveType(superclass);
			if(superClassNode == null)
					return children;
			
			List temp = (new ArrayList(Arrays.asList(children)));
			temp.addAll(Arrays.asList(superClassNode.getAllChildren(type)));
					
			return temp.toArray(children);
	}
	
	private ParseNode getAllChild(Class type, String name)
	{
		ParseNode child = getChild(type, name);
		if(child != null)
			return child;
			
		ClassNode superClassNode = resolveType(superclass);
		if(superClassNode == null)
			return superClassNode.getAllChild(type, name);
			
		return null;
	}

	/**
	 * Returns the <code>ClassNode</code> representing the superclass of the entity (class, interface, primitive type or void) represented by this <code>ClassNode</code>.
	 */
	public ClassNode getSuperclassNode()
	{
		return ClassNode.forName(getSuperclass(), this);
	}
	

	/**
	 * Determines if the class or interface represented by this <code>ClassNode</code> object is either the same as, or is a superclass or superinterface of, 
	 * the class or interface represented by the specified <code>ClassNode</code> parameter.
	 */
	public boolean isAssignableFrom(ClassNode klass)
	{
		if(klass == this)
			return true;
		
		ClassNode superNode = klass.getSuperclassNode();
		if(superNode == null)
				return false;
		else
				return isAssignableFrom(superNode);
	}

	/**
	 * Determines if the class or interface represented by this <code>ClassNode</code> object is either the same as, or is a superclass or superinterface of, 
	 * the class or interface named by the specified String parameter.
	 */
	public boolean isAssignableFrom(String klass)
	{
		return isAssignableFrom(resolveType(klass));
	}

	/**
	 * Returns an array containing <code>String</code> objects representing all the superclasses and superinterfaces 
	 * of the class represented by this <code>ClassNode</code> object.
	 */
	public String[] getSuperclasses()
	{
		List superclasses = new ArrayList();
		superclasses.add(superclass);
		ClassNode superClass = resolveType(superclass);
		if( superClass != null )
			superclasses.addAll(Arrays.asList(superClass.getSuperclasses()));
		return (String[]) superclasses.toArray(new String[0]);
	}

	/**
	 * Returns an array containing <code>String</code> objects representing all the subclasses and subinterfaces 
	 * of the class represented by this <code>ClassNode</code> object.
	 */
	public ClassNode[] getSubclasses()
	{
		List subclasses = new ArrayList();
		PackageNode[] packages = getRoot().getPackages();
		for(int i = 0; i < packages.length; i++)
		{
			ClassNode[] classes = packages[i].getClasses();
			for(int j = 0; j < classes.length; j++)
			{
				ClassNode klass = classes[j];
				if( this.isAssignableFrom(klass) && !klass.equals(this) )
					subclasses.add(klass);
			}
		}
		return (ClassNode[]) subclasses.toArray();
	}

	boolean isAddableTo(ChildBearingNode node)
	{
		return (node instanceof CodeBearingNode) ||
			   (node instanceof PackageNode) ||
			   (node instanceof ClassNode);
	}
		
	/**
	* Adds new members by describing them in source code.  For
	*  example,
	*
	* <pre>
	* aClass.extend(%{
	*	private int x;
	*	public void getX() {
	*		return x;
	* 	}
	* }%);
	* </pre>
	* would construct a new <code>FieldNode</code> and <code>MethodNode</code> for x and
	* <code>getX()</code>, and add them to <code>aClass</code>
	*
	* Accepts any code valid within the body of a class declaration.
	*/
    public void extend(String text)
	{
		ClassNode c = NodeFactory.createClass("class extension { " + text + "}");
		Iterator i = c.getChildren().iterator();
		while(i.hasNext())
			add((ParseNode)i.next());
	}
	

	/**
	 * Returns the <code>ClassNode</code> object associated with the class or interface with the given string name in the given
	 * context. The context parameter has two functions: first, it identifies a particular parse tree to search in.  For this
	 * purpose, it can be any node in that tree.  Second, it identifies a context in which ambiguous names will
	 * be resolved - for this purpose, the import statements of the class within which the context node was declared will be
	 * respected.
	 *
	 * Usually, the context parameter will simply be the target of the active transformation - for example,
	 * <pre>
	 * satisfies<methodDefinition>
	 * public addVisitMethod(ClassNode target, String visitor)
	 * {
	 * 	ClassNode visitorClass = ClassNode.forName(visitor, target);
	 *	...
	 *  }
	 * </pre>
	 */
	public static ClassNode forName(String name, ParseNode context)
	{
		ClassNode classContext;
		
		if(context == null)
			return null;
			
		if(context instanceof ClassNode && ((ClassNode)context).getDeclaringClass() == null)
			return ((ClassNode)context).resolveType(name);
			
		return forName(name, context.getAncestor(ClassNode.class));
	}

	//doesn't work with names of form package.Class.Inner	
	private ClassNode resolveType(String type)
	{
		int dot = type.lastIndexOf(".");
		if(dot != -1)
		{
			String pkg = type.substring(0, dot);
			String cls = type.substring(dot+1);
			PackageNode p = getRoot().getPackage(pkg);
			if(p != null)
			{
				ClassNode c = p.getClass(cls);
				if(c != null)
					return c;
			}
		}
		
		ClassNode c = getPackage().getClass(type);
		if(c != null)
			return c;
		
		c = getRoot().getDefaultPackage().getClass(type);
		if(c != null)
			return c;
		
		ImportNode[] imports = getImports();
		for(int i = 0; i < imports.length; i++)
		{
			String impt = imports[i].getName();
			int lastDot = impt.lastIndexOf(".");
			String preDot = impt.substring(0, lastDot);
			String postDot = impt.substring(lastDot + 1);
			
			if(postDot.equals("*") || postDot.equals(type))
			{
				PackageNode pkg = getRoot().getPackage(preDot);
				if(pkg != null)
				{
					c = pkg.getClass(type);
					if(c != null)
						return c;
				}
			}
		}
		
		return null;
	}
}