package ca.ubc.cs.elide.visitors;

import ca.ubc.cs.elide.nodes.*;
import java.util.*;
import java.io.*;
import java.lang.reflect.*;
import org.apache.oro.text.perl.*; 

public class Printer extends NodeVisitor
{
	private int tablevel = 0;
	protected Writer writer;
	
	public Printer(Writer writer) { this.writer = writer; }

	public void visitNode(BlockNode node)
	{
		append("/*start block*/");
		append(node.getCodeElements());
		append("/*end block*/");
	}
	
	public void visitNode(ClassNode node)
	{
		if(node.getParent() instanceof PackageNode)
		{
			append(node.getImports());
			appendln();
			appendln();
		}
				
		appendModifiers(node);
		
		if(node.isInterface())
			append(" interface ");
		else
			append(" class ");
		
		append(node.getShortName());
		
		if(! node.isInterface())
			append(" extends " + node.getSuperclass());
		
		if(node.getInterfaces().length > 0)
		{
			appendln();
			if(node.isInterface())
				append("\textends ");
			else
				append("\timplements ");
			append(node.getInterfaces(), ", ");
		}
		
		startBlock();
		
		append(node.getDeclaredFields());
		append(node.getStaticBlocks());
		append(node.getBlocks());
		append(node.getDeclaredMethods());
		append(node.getDeclaredClasses());
		
		endBlock();
	}
	
	public void visitNode(FieldNode node)
	{
		appendModifiers(node);
		append(" " + node.getType() + " " + node.getName());
		if(node.getInitializer() != null)
		{
			append(" = ");
			append(node.getInitializer());
		}
		
		append(";");
	}
	
	public void visitNode(ImportNode node)
	{
		append("import " + node.getName() + ";");
	}
	
	public void visitNode(InstanceBlockNode node)
	{
		startBlock();
		append(node.getCodeElements());
		endBlock();
	}
	
	public void visitNode(MethodNode node)
	{
		appendModifiers(node);
		append(node.getReturnType() + " " + node.getName());
		append("(");
		append(node.getParameters(), ", ");
		append(")");
		
		if(node.getExceptionTypes().length > 0)
		{
			appendln();
			append("\tthrows ");
			append(node.getExceptionTypes(), ", ");
		}
		
		if(!node.isBodyless())
		{
			startBlock();
			append(node.getCodeElements());
			endBlock();
		}
		else
			append(";");
	}
		
	public void visitNode(ParameterNode node)
	{
		appendModifiers(node);
		append(node.getType());
		append(" ");
		append(node.getName());
	}
	
	public void visitNode(ParseNode node)
	{
		append(node.getName());
	}
	
	public void visitNode(PackageNode node)
	{
		append("package " + node.getName() + ";\n");
		append(node.getClasses());
	}

	public void visitNode(StaticBlockNode node)
	{
		append("static");
		startBlock();
		append(node.getCodeElements());
		endBlock();
	}
	
	public void visitNode(TextNode node)
	{
		append(node.getText());
	}
	
	public void visitNode(TransformNode node)
	{
		//this doesn't work.  why not?
		append("/**" + node.toString() + "**/ ");
	}
	
	protected void appendModifiers(ModifierBearingNode node)
	{
		if(!node.getComment().equals(""))
			append(node.getComment());
		
		append(Modifier.toString(node.getModifiers()));
		if( node.getModifiers() != 0 ) append(" ");
	}
		
	protected void startBlock()
	{
		appendln();
		append("{");
		tablevel++;
		appendln();
	}
	
	protected void endBlock()
	{
		tablevel--;
		appendln();
		append("}");
	}
	
	protected void appendln()
	{
		append("\n");
	}
	
	protected void append(Object obj)
	{
		if(obj instanceof ParseNode)
			visit((ParseNode)obj);
		else
		{
			String text = obj.toString();
			Perl5Util perl = new Perl5Util();
			
			StringBuffer tabs = new StringBuffer();
			for(int i = 0; i < tablevel; i++)
					tabs.append("\t");

			write(perl.substitute("s/\\n/\n" + tabs + "/g", text));
		}
	}
	
	protected void append(Object[] list)
	{
		append(list, "\n");
		appendln();
	}
	
	protected void append(Object[] list, String separator)
	{
		int i = 0;
		while(i < list.length)
		{
			append(list[i]);
			if(++i < list.length )
				append(separator);
		}
	}
	
	public void flush()
	{
		try { writer.flush(); }
		catch(IOException e) { e.printStackTrace(); }
	}
	
	private void write(String str)
	{
		try { writer.write(str); }
		catch(IOException e) { e.printStackTrace(); }
	}
}